#include <stdio.h>
#include <stdlib.h>
#include <EzGraph.h>
#include "config.h"
#include "layout.h"
#include "param.h"
#include "draw.h"
#include "sprite.h"
#include "obj.h"
#include "obj_draw.h"
#include "coordinate.h"
#include "rnd.h"

void draw_screen(void);
void draw_map(void);
void draw_radar(void);
void draw_bars(void);

pos2_t stars_pos[2][3][STARS_NUM];

void draw_init(void){
    int i, j, k;
    sprite_init();
    obj_draw_init();
    
    srnd(1);
    for (i = 0 ; i < 2 ; i++){
        for (j = 0 ; j < 3 ; j++){
            for (k = 0 ; k < STARS_NUM ; k++){
                stars_pos[i][j][k].x = RRnd(0, SCREEN_WIDTH-1);
                stars_pos[i][j][k].y = RRnd(0, SCREEN_HEIGHT-1);
            }
        }
    }
}

#ifdef DEBUG
int wsensor_enabled = 0;
void enable_wsensor(void){ wsensor_enabled = 1; puts("hoge");}
void disable_wsensor(void){ wsensor_enabled = 0; }
int is_wsensor_enabled(void){ return wsensor_enabled;}
#endif



void draw(void){
#ifdef DEBUG
    if (wsensor_enabled){
        draw_wsensor_objs();
    }
#endif
    draw_screen();
    draw_map();
    draw_radar();
    draw_bars();
    //    draw_obj(100, 100, gi_chee);
    //    draw_obj(SCREEN_OX-16, SCREEN_OY-16, gi_chee_back);
    EzShowBuffer();
}

void draw_stars(void){
    int k;
    dir_t dir;

    dir = get_myrobo_dir();
    EzSetColorByRGB(160, 160, 160);
    for (k = 0 ; k < STARS_NUM ; k++){
        EzDrawPointB(stars_pos[dir][0][k].x, stars_pos[dir][0][k].y);
    }
    EzSetColorByRGB(200, 200, 200);
    for (k = 0 ; k < STARS_NUM ; k++){
        EzFillBoxB(stars_pos[dir][1][k].x, stars_pos[dir][1][k].y,
                   stars_pos[dir][1][k].x+1, stars_pos[dir][1][k].y+1);
    }
    EzSetColorByRGB(255, 255, 255);
    for (k = 0 ; k < STARS_NUM ; k++){
        EzFillBoxB(stars_pos[dir][2][k].x, stars_pos[dir][2][k].y,
                   stars_pos[dir][2][k].x+2, stars_pos[dir][2][k].y+2);
    }
}

void draw_screen(void){
    draw_stars();
    draw_screen_objs();
}

void draw_map(void){
    EzSetColorByRGB(0, 0, 80);
    EzFillBoxB(MAP_LTX, MAP_LTY, MAP_RBX, MAP_RBY);
    //    EzFillBoxB(SCREEN_WIDTH+1, 0, WIN_WIDTH-1, WIN_HEIGHT-1);
    draw_map_objs();
    EzSetColorByRGB(255, 255, 255);
    EzDrawBoxB(MAP_LTX-1, MAP_LTY-1, MAP_RBX+1, MAP_RBY+1);
}

void draw_radar(void){
    EzSetColorByRGB(0, 0, 80);
    EzFillCircleB(RADAR_F_X, RADAR_F_Y, RADAR_R);
    EzFillCircleB(RADAR_B_X, RADAR_B_Y, RADAR_R);
    draw_radar_objs();
    EzSetColorByRGB(255, 255, 255);
    EzDrawCircleB(RADAR_F_X, RADAR_F_Y, RADAR_R);
    EzDrawCircleB(RADAR_B_X, RADAR_B_Y, RADAR_R);
    EzDrawCircleB(RADAR_F_X, RADAR_F_Y, RADAR_IR);
    EzDrawCircleB(RADAR_B_X, RADAR_B_Y, RADAR_IR);
    EzDrawLineB(RADAR_F_X-RADAR_IR, RADAR_F_Y-RADAR_IR,
                RADAR_F_X+RADAR_IR, RADAR_F_Y+RADAR_IR);
    EzDrawLineB(RADAR_F_X+RADAR_IR, RADAR_F_Y-RADAR_IR,
                RADAR_F_X-RADAR_IR, RADAR_F_Y+RADAR_IR);
    EzDrawLineB(RADAR_B_X-RADAR_IR, RADAR_B_Y-RADAR_IR,
                RADAR_B_X+RADAR_IR, RADAR_B_Y+RADAR_IR);
    EzDrawLineB(RADAR_B_X+RADAR_IR, RADAR_B_Y-RADAR_IR,
                RADAR_B_X-RADAR_IR, RADAR_B_Y+RADAR_IR);
    //    EzDrawBoxB(RADAR_F_LTX-1, RADAR_F_LTY-1, RADAR_F_RBX+1, RADAR_F_RBY+1);
    //    EzDrawBoxB(RADAR_B_LTX-1, RADAR_B_LTY-1, RADAR_B_RBX+1, RADAR_B_RBY+1);
}

void draw_bars(void){
    int hp, speed;

    EzSetColorByRGB(255, 255, 255);
    EzDrawBoxB(HP_BAR_LTX-1, HP_BAR_LTY-1, HP_BAR_RBX+1, HP_BAR_RBY+1);
    EzDrawBoxB(SPEED_BAR_LTX-1, SPEED_BAR_LTY-1,
               SPEED_BAR_RBX+1, SPEED_BAR_RBY+1);

    hp = get_myrobo_hp();
    EzSetColorByRGB(0, 255, 0);
    if (hp > 0){
        EzFillBoxB(HP_BAR_LTX, HP_BAR_LTY,
                   HP_BAR_LTX+hp*HP_BAR_WIDTH/MYROBO_HP, HP_BAR_RBY);
    }

    speed = (int)get_myrobo_velo_z();
    //    printf("speed:%d\n", speed);
    if (get_myrobo_dir() == di_b) speed = -speed;
    EzSetColorByRGB(0, 255, 255);
    if (speed > 0){
        EzFillBoxB(SPEED_BAR_BASE,
                   SPEED_BAR_LTY,
                   SPEED_BAR_BASE + speed*SPEED_BAR_WIDTH/2/MYROBO_FVELO_MAX,
                   SPEED_BAR_RBY);
    }
    else if (speed < 0){
        EzFillBoxB(SPEED_BAR_BASE + speed*SPEED_BAR_WIDTH/2/MYROBO_FVELO_MAX,
                   SPEED_BAR_LTY,
                   SPEED_BAR_BASE,
                   SPEED_BAR_RBY);
    }
    else {
        EzDrawLineB(SPEED_BAR_BASE, 
                    SPEED_BAR_LTY,
                    SPEED_BAR_BASE,
                    SPEED_BAR_RBY);
    }
}
