#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <EzGraph.h>
#include "config.h"
#include "param.h"
#include "coordinate.h"
#include "obj.h"
#include "drone.h"
#include "sprite.h"
#include "rnd.h"

static sprite_id_t drone_sids[] = {
    si_drone_12,
    si_drone_13,
    si_drone_14,
    si_drone_15,
    si_drone_00,
    si_drone_01,
    si_drone_02,
    si_drone_03,
    si_drone_04,
    si_drone_05,
    si_drone_06,
    si_drone_07,
    si_drone_08,
    si_drone_09,
    si_drone_10,
    si_drone_11,
};

static ivec2_t drone_rot_velo[MYROBO_TURN_FRAME_NUM];

void drone_fire(obj_t *drone, dir_t dir);

void drone_init(void){
    int i;
    double r;
    for (i = 0 ; i < 16 ; i++){
        r = Deg2Rad(360*i/16);
        drone_rot_velo[i].x = (int)(cos(r) * DRONE_VELO_MAX);
        drone_rot_velo[i].y = (int)(sin(r) * DRONE_VELO_MAX);
        //        printf("velo %d (%d, %d)\n", i, turn_velo[i].x, turn_velo[i].y);
    }
}




void setup_drone(obj_t *obj){
    obj->drone.rot = 11;
    obj->drone.mode = dm_move_xy;
    obj->drone.turn_count = 0;
    obj->drone.fire_count = 0;
}


void rot_drone_to(obj_t *drone, int rot){
    drone->drone.turn_count ++;
    if (drone->drone.turn_count >= DRONE_TURN_SPEED){
        if (drone->drone.rot != rot){
            drone->drone.rot = (drone->drone.rot+1) % 16;
        }
        drone->drone.turn_count = 0;
    }
}

void act_drone(obj_t *obj){
    //    printf("drone rot:%d mode:%d\n", obj->drone.rot, obj->drone.mode);
    // zǼΥƤȤߤ
    if (abs(obj->rpos3.z) > RADAR_RANGE) return;

    switch (obj->drone.mode){
    case dm_move_xy:
        if (obj->rpos3.x < -8){
            rot_drone_to(obj, 0);
        }
        else if (obj->rpos3.x > 8){
            rot_drone_to(obj, 8);
        }
        else {
            puts("check mode to fb");
            if (abs(obj->rpos3.y) < 8 && rrnd(0, 9) < 5){
                obj->drone.mode = dm_turn_fb;
            }
        }
        if (obj->rpos3.y > 0){
            obj->velo3.y -= (-obj->velo3.y >= DRONE_VELO_MAX/2) ? 0 : 1;
        }
        else if (obj->rpos3.y < 0){
            obj->velo3.y += (obj->velo3.y >= DRONE_VELO_MAX/2) ? 0 : 1;
        }
        break;
    case dm_turn_fb:
        obj->velo3.y = 0;
        if (obj->rpos3.z < 0){
            rot_drone_to(obj, 4);
            if (obj->drone.rot == 4){
                if (rrnd(0, 9) < 2) obj->drone.mode = dm_move_z;
            }
        }
        else if (obj->rpos3.z > 0){
            rot_drone_to(obj, 12);
            if (obj->drone.rot == 12){
                if (rrnd(0, 9) < 2) obj->drone.mode = dm_move_z;
            }
        }
    case dm_move_z:
        if (obj->drone.rot == 4){
            if (obj->rpos3.z > 128){
                if (rrnd(0, 9) < 2){
                    obj->drone.mode = dm_move_xy;
                }
            }
        }
        else if (obj->drone.rot == 12){
            if (obj->rpos3.z < -64){
                if (rrnd(0, 9) < 2){
                    obj->drone.mode = dm_move_xy;
                }
            }
        }
        break;
    default:
        break;
    }

     
#ifdef DUMMY
    //    obj->drone.turn_count ++;
    if (obj->drone.turn_count == 12){
        obj->drone.rot ++;
        if (obj->drone.rot > 15) obj->drone.rot = 0;
        obj->drone.turn_count = 0;
    }
#endif
    obj->sid = drone_sids[obj->drone.rot];
    //    printf("drone rot:%d (%d, %d)\n", obj->drone.rot, obj->rpos3.x, obj->rpos3.y);

    obj->velo3.x = drone_rot_velo[obj->drone.rot].x;
    obj->velo3.z = drone_rot_velo[obj->drone.rot].y;

    if (obj->drone.fire_count > 0){
        obj->drone.fire_count --;
    }
    if (obj->drone.rot == 4 && obj->rpos3.z < 0){
        drone_fire(obj, di_f);
    }
    else if (obj->drone.rot == 12 && obj->rpos3.z > 0){
        drone_fire(obj, di_b);
    }
    move_obj(obj);
}


void drone_fire(obj_t *drone, dir_t dir){
    obj_t *obj;
    pos3_t pos3 = drone->pos3;

    if (drone->drone.fire_count > 0) return;
    pos3.z += (dir == di_f ? drone->size/4 : -drone->size/4);
    obj = create_obj(ot_enemybullet, pos3);
    set_dir_enemybullet(obj, dir);
    drone->drone.fire_count = DRONE_FIRE_SPEED;
}

