#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <EzGraph.h>
#include "config.h"
#include "event.h"
#include "event_in.h"
#include "obj.h"
#include "file.h"


typedef struct {
    event_trigger_type_t et;
    char *name;
} event_trigger_info_t;
typedef struct {
    event_op_type_t eo;
    char *name;
} event_op_info_t;

#define ETInfo(name) {et_ ## name, #name}
event_trigger_info_t event_trigger_infos[] = {
    ETInfo(start),
    ETInfo(complete),
    ETInfo(fail),
    ETInfo(die),
    ETInfo(in),
    ETInfo(out),
    ETInfo(time),
};
#define EVENT_TRIGGER_INFO_NUM \
    (sizeof(event_trigger_infos)/sizeof(event_trigger_info_t))
#define EOInfo(name) {eo_ ## name, #name}
event_op_info_t event_op_infos[] = {
    EOInfo(create),
    EOInfo(create_region),
    EOInfo(delete),
    EOInfo(destroy),
};
#define EVENT_OP_INFO_NUM \
    (sizeof(event_op_infos)/sizeof(event_op_info_t))


static event_list_t event_list;

event_id_t enqueue_event(event_trigger_type_t et, event_op_type_t eo);


void event_init(void){
}

void reset_event_list(void){
    event_list.num = 0;
}

event_trigger_type_t event_trigger_name2type(char *str){
    int i;
    for (i = 0 ; i < EVENT_TRIGGER_INFO_NUM ; i++){
        if (strcmp(str, event_trigger_infos[i].name) == 0){
            return event_trigger_infos[i].et;
        }
    }
    return et_invalid;
}
event_op_type_t event_op_name2type(char *str){
    int i;
    for (i = 0 ; i < EVENT_OP_INFO_NUM ; i++){
        if (strcmp(str, event_op_infos[i].name) == 0){
            return event_op_infos[i].eo;
        }
    }
    return eo_invalid;
}

event_id_t enqueue_event(event_trigger_type_t et, event_op_type_t eo){
    if (event_list.num >= MISSION_EVENT_NUM_MAX){
        log_error("event queue overflow\n");
        exit(1);
    }
    event_list.events[event_list.num].trigger.type = et;
    event_list.events[event_list.num].op.type = eo;
    event_list.num ++;
    return event_list.num-1;
}

event_id_t enqueue_event_create(event_trigger_type_t et,
                                obj_type_t obj_type, pos3_t pos3)
{
    event_id_t eid;
    event_t *event;
    
    eid = enqueue_event(et, eo_create);
    event = &event_list.events[eid];
    
    event->op.param.create.obj_type = obj_type;
    event->op.param.create.pos3 = pos3;

    return eid;
}

event_id_t enqueue_event_create_region(event_trigger_type_t et,
                                       obj_type_t obj_type, region3_t region3,
                                       int num)
{
    event_id_t eid;
    event_t *event;
    
    eid = enqueue_event(et, eo_create_region);
    event = &event_list.events[eid];
    
    event->op.param.create_region.obj_type = obj_type;
    event->op.param.create_region.region3 = region3;
    event->op.param.create_region.num = num;
    return eid;
}

void register_events(void){
    event_id_t eid = 0;
    event_t *event;

    for (eid = 0 ; eid < event_list.num ; eid++){
        printf("register event %d\n", eid);
        event = &event_list.events[eid];
        switch (event->op.type){
        case eo_create:
            create_obj(event->op.param.create.obj_type, 
                       event->op.param.create.pos3);
            break;
        case eo_create_region:
            create_obj_region(event->op.param.create_region.obj_type, 
                              event->op.param.create_region.region3,
                              event->op.param.create_region.num);
            break;
        default:
            assert(0);
        }
    }
    //    exit(1);
}


#ifdef DUMMY
EventList_t create_event_list(void){
    event_list_t *event_list;
    event_list = malloc(sizeof(event_list_t));
    if (event_list == NULL){
        printf("failed create_event_array\n");
        exit(1);
    }
    event_list->num = 0;
    return event_list;
}

Event_t create_event(EventList_t event_list){
    assert (event_list->num < MISSION_EVENT_NUM_MAX);
    event_list->num ++;
    return &(event_list->events[event_list->num-1]);
}

void set_event_create(Event_t event, obj_type_t obj_type, pos3_t pos3){
    event_trigger_t trigger;
    event_op_t op;

    trigger.type = et_started;
    op.type = eo_create;
    op.param.create.obj_type = obj_type;
    op.param.create.pos3 = pos3;

    event->trigger = trigger;
    event->op = op;
}

void set_event_create_region(Event_t event, obj_type_t obj_type, 
                             region3_t region3, int num)
{
    event_trigger_t trigger;
    event_op_t op;

    trigger.type = et_started;
    op.type = eo_create_region;
    op.param.create_region.obj_type = obj_type;
    op.param.create_region.region3 = region3;
    op.param.create_region.num = num;

    event->trigger = trigger;
    event->op = op;
}

void register_events(EventList_t event_list){
    event_id_t eid = 0;
    event_t *event;

    for (eid = 0 ; eid < event_list->num ; eid++){
        printf("register event %d\n", eid);
        event = &event_list->events[eid];
        switch (event->op.type){
        case eo_create:
            create_obj(event->op.param.create.obj_type, 
                       event->op.param.create.pos3);
            break;
        case eo_create_region:
            create_obj_region(event->op.param.create_region.obj_type, 
                              event->op.param.create_region.region3,
                              event->op.param.create_region.num);
            break;
        default:
            assert(0);
        }
    }
    //    exit(1);
}

#endif

#ifdef DUMMY

//[MISSION_EVENT_NUM_MAX]



#endif




#ifdef DUMMY
    obj->pos3.x = RRnd(0, SPACE_XSIZE);
    obj->pos3.y = RRnd(0, SPACE_YSIZE);
    obj->pos3.z = RRnd(0, SPACE_ZSIZE);

        obj->pos3.x = SPACE_XSIZE/2;
        obj->pos3.y = SPACE_YSIZE/2-8;
    obj->pos3.z = SPACE_ZSIZE/15;


            obj->pos3.x = tmp%5*SPACE_XSIZE/4;
            //    obj->pos3.x = tmp%5*obj->size;
    obj->pos3.y = SPACE_YSIZE/2;
    //    obj->pos3.z = (tmp)/5 * SPACE_ZSIZE/64+SPACE_ZSIZE/15;
    obj->pos3.z = (tmp)/5 * SPACE_ZSIZE/256+SPACE_ZSIZE/15;
    tmp ++;

    /*
    */
    /*
    obj->pos3.x = RRnd(0, SPACE_XSIZE);
    obj->pos3.y = RRnd(0, SPACE_YSIZE);
    obj->pos3.z = RRnd(0, SPACE_ZSIZE);
    */
#endif
