#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <EzGraph.h>
#include "config.h"
#include "mission.h"
#include "coordinate.h"
#include "obj.h"
#include "file.h"
#include "jstr.h"

#define BUF_SIZE 1024

static char dpath[BUF_SIZE];
static char fpath[BUF_SIZE];
static int line_no;
static mission_t current_mission;
static int current_mission_no;

typedef struct {
    char title[JBufSize(MISSION_TITLE_LEN_MAX)];
    int top_score;
} mission_stat_t;


static char *mission_names[] = {
    "training_move",
    "training_fire",
    "dogfight",
    "scout",
    "mine",
    "dogfight2",
    "destroyer",
    "attack",
    "fleet_attack",

    //    "test",
    //    "test2",
    //    "training",
    //    "empty_space",
};
#define MISSION_NUM (sizeof(mission_names)/sizeof(char*))

mission_stat_t mission_stats[MISSION_NUM];


typedef enum {
    ct_invalid = -1,
    ct_begin = 0,
    ct_title=ct_begin,
    ct_briefing,
    ct_event,
    ct_end
} cmd_type_t;

static char *cmd_strs[] = {
    "title",
    "briefing",
    "event",
};

static cmd_type_t current_ct;
static event_trigger_type_t current_et;

cmd_type_t cmd_str2type(char *str);
int load_mission_title(int mission_no);
int load_mission_title_pre_callback(void);
int load_mission_title_callback(char *line);
int load_mission_title_pre_callback(void);
int load_mission_title_post_callback(void);
int load_mission_callback(char *line);
int load_mission_pre_callback(void);
int load_mission_post_callback(void);
int read_event(event_op_type_t eo, char *line);
obj_type_t read_objective(char **line);
pos3_t read_pos3(char **line);
region3_t read_region3(char **line);
int read_int(char **line);
int read_str(char **line, char *buf);
int read_mission_file(int mission_no,
                      int(*pre_func)(void),
                      int(*func)(char*),
                      int(*post_func)(void));


#define SkipSpace(ptr) {while(*(ptr) ==' '){(ptr) ++;}}
#define SkipToSpace(ptr) {while(*(ptr) !=' ' && *(ptr) != '\0'){(ptr) ++;}}
// chr¸ߤ('\0'chrи)Τ߻Ѳǽ
#define SkipToChr(ptr, chr) {while(*(ptr) != (chr)){(ptr) ++;}}
#define NextChr() ((*line)++)
void chop_line(char **line);



void mission_init(char *argv0){
	// Visual StudioǤγȯϡ
	// ¹ԥե뤬Ȱۤʤǥ쥯ȥ˺Τǡ
	// ¹ԥե뤫Хǥ쥯ȥǤưʤ
	// ȥǥ쥯ȥϥ(ץȥե?)ξʤΤǡ
	// ȤꤢȤлˤƤ
#if defined(_MSC_VER) && defined(_DEBUG)
    sprintf(dpath, "%s", MISSION_DIR);
#else
    sprintf(dpath, "%s/%s", get_exe_dpath(argv0), MISSION_DIR);
    printf("dpath:%s\n", dpath);
#endif
}

int get_mission_num(void){
    return MISSION_NUM;
}
char *get_mission_name(int no){
    assert(no >= 0);
    assert(no < MISSION_NUM);
    return mission_names[no];
}
char *get_mission_title(int no){
    assert(no >= 0);
    assert(no < MISSION_NUM);
    return mission_stats[no].title;
}

void set_current_mission_no(int no){
    assert(no >= 0);
    assert(no < MISSION_NUM);
    current_mission_no = no;
}
int get_current_mission_no(void){
    return current_mission_no;
}
char *get_current_mission_title(void){
    return get_mission_title(current_mission_no);
}
char *get_current_mission_briefing(int line_no){
    return current_mission.briefing[line_no];
}
int get_current_mission_briefing_num(void){
    return current_mission.briefing_num;
}

int load_mission_titles(void){
    int i;

    for (i = 0 ; i < MISSION_NUM ; i++){
        load_mission_title(i);
    }
    return 0;
}
int load_mission_title(int mission_no){
    return read_mission_file(mission_no,
                             load_mission_title_pre_callback,
                             load_mission_title_callback,
                             load_mission_title_post_callback);
}


int load_mission_title_pre_callback(void){
    mission_stats[current_mission_no].title[0] = '\0';
    return 0;
}
int load_mission_title_callback(char *line){
    if (current_ct == ct_title){
        strcpy(mission_stats[current_mission_no].title, line);
        return 1;
    }
    if (strcmp(line, "@title") == 0) current_ct = ct_title;
    return 0;
}
int load_mission_title_post_callback(void){
    if (mission_stats[current_mission_no].title[0] == '\0'){
        log_error("Title not found in mission file %s.\n", fpath);
        exit(1);
    }
    return 0;
}

int load_mission(void){
    reset_event_list();
    current_mission.briefing_num = 0;
    return read_mission_file(current_mission_no,
                             load_mission_pre_callback,
                             load_mission_callback,
                             load_mission_post_callback);
}
int load_mission_pre_callback(void){
    return 0;
}
int load_mission_callback(char *line){
    char *cmd_str;
    char buf[BUF_SIZE];

    if (line[0] == '@'){
        cmd_str = &line[1];
        current_ct = cmd_str2type(cmd_str);
        if (current_ct == ct_invalid){
            log_error("Unknown command '%s'"
                      " at line %d in '%s'\n",
                      cmd_str, line_no, fpath);
                exit(1);
        }
    }
    else{
		char c;
        switch (current_ct){
            case ct_title:
                // titlemission_statsˤΤפɤ߹
                strcpy(current_mission.title, line);
                break;
            case ct_briefing:
                strcpy(current_mission.briefing[current_mission.briefing_num],
                       line);
                current_mission.briefing_num ++;
                break;
            case ct_event:
				// @note "%s {"Ǥ'{'³ʤƤ%sɤ߹ߤ
                // 1֤Τǡä%cɤǤ'{'å
                if (sscanf(line, "%s %c", buf, &c) == 2 && c == '{'){
                    current_et = event_trigger_name2type(buf);
                    if (current_et == et_invalid){
                        log_error("Unknown event trigger '%s'"
                                  " at line %d in '%s'\n",
                                  buf, line_no, fpath);
                        exit(1);
                    }
                }
                else if (line[0] == '}'){
                    current_et = et_invalid;
                }
                else if (read_str(&line, buf) == 0){
                    event_op_type_t eo;
                    eo = event_op_name2type(buf);
                    if (eo == eo_invalid){
                        log_error("Unknown event op '%s'"
                                  " at line %d in '%s'\n",
                                  buf, line_no, fpath);
                        exit(1);
                    }
                    if (read_event(eo, line)){
                        log_error("Illegal event args '%s'"
                                  " at line %d in '%s'\n",
                                  line, line_no, fpath);
                        exit(1);
                    }
                }
                else {
                    log_error("Illegal line '%s'"
                              " at line %d in '%s'\n",
                              line, line_no, fpath);
                    exit(1);
                }
                break;
            default:
                assert(0);
                break;
        }
    }
    return 0;
}

int load_mission_post_callback(void){
    return 0;
}

int read_event(event_op_type_t eo, char *line){
    obj_type_t ot;
    pos3_t pos3;
    region3_t region3;
    int num;

    switch (eo){
    case eo_create:
        ot = read_objective(&line);
        pos3 = read_pos3(&line);
        enqueue_event_create(current_et, ot, pos3);
        break;
    case eo_create_region:
        ot = read_objective(&line);
        region3 = read_region3(&line);
        num = read_int(&line);
        enqueue_event_create_region(current_et, ot, region3, num);
        break;
    default:
        assert(0);
        break;
    }
    return 0;
}
obj_type_t read_objective(char **line){
    obj_type_t ot;
    char buf[BUF_SIZE];
    if (read_str(line, buf) != 0){
        log_error("Objective missing (line:%s)"
                  " at line %d in '%s'\n",
                  line, line_no, fpath);
        exit(1);
    }
    ot = obj_name2type(buf);
    if (ot == ot_invalid){
        log_error("Unknown objective '%s'"
                  " at line %d in '%s'\n",
                  buf, line_no, fpath);
        exit(1);
    }
    return ot;
}

pos3_t read_pos3(char **line){
    pos3_t pos3;
//    double x, y, z;
    int x, y, z;

//    if (sscanf(*line, "(%f%%, %f%%, %f%%)", &x, &y, &z) == 3){
    if (sscanf(*line, "(%d%%, %d%%, %d%%)", &x, &y, &z) == 3){
        pos3.x = (int)(SPACE_XSIZE * x / 100);
        pos3.y = (int)(SPACE_YSIZE * y / 100);
        pos3.z = (int)(SPACE_ZSIZE * z / 100);
    }
    else {
        log_error("Illegal pos '%s'"
                  " at line %d in '%s'\n",
                  *line, line_no, fpath);
        exit(1);
    }
    SkipToChr(*line, ')');
    NextChr();
    return pos3;
}

region3_t read_region3(char **line){
    region3_t region3;
    char c;
//    double x, y, z;
//    int sx, sy, sz, ex, ey, ez;


//    if (sscanf(*line, "(%f%%, %f%%, %f%%)", &x, &y, &z) == 3){
    region3.st = read_pos3(line);
    //    SkipSpace(*line);
    //    SkipToSpace(*line);
    if (sscanf(*line, " %c ", &c) != 1 || c != '-'){
        log_error("Illegal range '%c'"
                  " at line %d in '%s'\n",
                  c, line_no, fpath);
        exit(1);
    }
    SkipToChr(*line, '-');
    NextChr();
    region3.en = read_pos3(line);
    return region3;
}

int read_int(char **line){
    int val;

    if (sscanf(*line, "%d", &val) != 1){
        log_error("Illegal int val '%s'"
                  " at line %d in '%s'\n",
                  *line, line_no, fpath);
        exit(1);
    }
    SkipToSpace(*line);
    return val;
}

int read_str(char **line, char *buf){
    if (sscanf(*line, "%s", buf) != 1){
        return -1;
    }
    *line += strlen(buf);
    SkipSpace(*line);
    return 0;
}


int read_mission_file(int mission_no,
                      int(*pre_func)(void),
                      int(*func)(char*),
                      int(*post_func)(void))
{
    FILE *fp;
    char line_buf[BUF_SIZE];
    char *line;

    current_ct = ct_invalid;
    current_mission_no = mission_no;
    
    sprintf(fpath, "%s/%s.%s",
            dpath, mission_names[mission_no], MISSION_EXT);

    if (pre_func != NULL){
        if (pre_func()) return 1;
    }

    fp = fopen_bomcut(fpath, "r");
    line_no = 0;
    
    if (fp == NULL){
        log_error("Cannot open mission file %s.\n", fpath);
        exit(1);
    }
    while (fgets(line_buf, BUF_SIZE, fp) != NULL){
        line_no ++;

        // Ƭ#ԤɤФ
        line = line_buf;
        chop_line(&line);
        if (line[0] == '#') continue;
        if (current_ct != ct_briefing){
            if (line[0] == '\0') continue;
        }
        // Хåؿ0֤顢ǽλ
        // @note ߥåե˥顼äץλ
        // ΤǡХåؿ֤ͤ0:/0:ʹߤɤ߹
        // 2Τ
        if (func(line)) break;
    }
    if (post_func != NULL){
        return post_func();
    }
    return 0;
}

// Ƭβԡ
// @note ƬˤĤƤϥݥline򤺤餹ȤбƤ
void chop_line(char **line){
    int last;

    // @todo ֥꡼ե󥰥ƥȤϹƬζĤ
    // (ܸʸޤǽ⤢ΤɤФʤjstrؿɬ)
    if (current_ct != ct_briefing){
        SkipSpace(*line);
    }
    last = strlen(*line)-1;
    while (last >= 0 && ((*line)[last] == '\n' || (*line)[last] == ' ')){
        last --;
    }
    (*line)[last+1] = '\0';
    
}

cmd_type_t cmd_str2type(char *str){
    cmd_type_t ct;

    for (ct = ct_begin ; ct < ct_end ; ct++){
        if (strcmp(str, cmd_strs[ct]) == 0){
            return ct;
        }
    }
    return ct_invalid;
}

void setup_mission(void){
    distribute_fragments();
    register_events();
}


#ifdef DUMMY
// ߥåϵư˽Τǡߥå˽դ­Ƥ
mission_id_t create_mission(char *name){
    //    if (mission_num >= MISSION_NUM_MAX) return INVALID_MISSION_ID;
    strcpy(missions[mission_num].name, name);
    missions[mission_num].event_list = create_event_list();
    current_mission = &missions[mission_num];
    return mission_num++;
}

Event_t create_event_create(obj_type_t obj_type, pos3_t pos3){
    Event_t event;

    event = create_event(current_mission->event_list);
    set_event_create(event, obj_type, pos3);
    return event;
}

Event_t create_event_create_region(obj_type_t obj_type, region3_t region3,
                                   int num)
{
    Event_t event;

    event = create_event(current_mission->event_list);
    set_event_create_region(event, obj_type, region3, num);
    return event;
}





#endif

#ifdef DUMMY

void register_events(void){
    int i;
    event_t *event;

    for (i = 0 ; i < event_num ; i++){
        event = &mission.events[i];
        if (event.trigger.type == et_start){
            printf("exec_event");
            exec_event(event);
        }
    }
}
#endif
