#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <EzGraph.h>
#include "config.h"
#include "param.h"
#include "uty.h"
#include "obj.h"
#include "myrobo.h"
#include "main_state.h"
#include "coordinate.h"

static obj_t *myrobo;

static double turn_degs[] = {
    0,
    1.23,
    4.93,
    11.12,
    19.73,
    30.60,
    43.48,
    58.01,
    73.70,
    90
};

static ivec2_t turn_velo[MYROBO_TURN_FRAME_NUM];

void set_myrobo_acc_x(int val);
void set_myrobo_acc_y(int val);
void set_myrobo_acc_z(int val);

void myrobo_init(void){
    int i;
    double r;
    for (i = 0 ; i < MYROBO_TURN_FRAME_NUM ; i++){
        if (i<=9){
            r = Deg2Rad(turn_degs[i]);
        }
        else {
            r = Deg2Rad(180-turn_degs[MYROBO_TURN_FRAME_NUM-i]);
        }
        turn_velo[i+1].x = (int)(sin(r) * MYROBO_FVELO_MAX);
        turn_velo[i+1].y = (int)(cos(r) * MYROBO_FVELO_MAX);
        printf("velo %d (%d, %d)\n", i, turn_velo[i].x, turn_velo[i].y);
    }
}

void setup_myrobo(obj_t *obj){
    myrobo = obj;

    obj->velo3.x = 0;
    obj->velo3.y = 0;
    obj->velo3.z = 0;
    obj->acc3.x = 0;
    obj->acc3.y = 0;
    obj->acc3.z = 0;

    obj->myrobo.dir = di_f;
    obj->myrobo.turn_count = 0;
    obj->myrobo.die_count = 0;
    obj->myrobo.zthrust = 0;
    obj->myrobo.is_firing = 0;
}

void get_myrobo_pos3(pos3_t *pos3){
    *pos3 = myrobo->pos3;
}


void myrobo_thrust_left(void){
    if (myrobo->myrobo.dir == di_f){
        set_myrobo_acc_x(-MYROBO_ACC);
    }
    else {
        set_myrobo_acc_x(MYROBO_ACC);
    }
    //    printf("tl acc:%f velo:%f\n", myrobo->acc3.x, myrobo->velo3.x);
}
void myrobo_thrust_right(void){
    if (myrobo->myrobo.dir == di_f){
        set_myrobo_acc_x(MYROBO_ACC);
    }
    else {
        set_myrobo_acc_x(-MYROBO_ACC);
    }
    //    printf("tr acc:%f velo:%f\n", myrobo->acc3.x, myrobo->velo3.x);
}
void myrobo_thrust_up(void){
    set_myrobo_acc_y(-MYROBO_ACC);
    //    printf("tl acc:%f velo:%f\n", myrobo->acc3.x, myrobo->velo3.x);
}
void myrobo_thrust_down(void){
    set_myrobo_acc_y(MYROBO_ACC);
    //    printf("tr acc:%f velo:%f\n", myrobo->acc3.x, myrobo->velo3.x);
}
void myrobo_thrust_forward(void){
    if (myrobo->myrobo.dir == di_f){
        set_myrobo_acc_z(MYROBO_ACC);
    }
    else {
        set_myrobo_acc_z(-MYROBO_ACC);
    }
}
void myrobo_thrust_backward(void){
    if (myrobo->myrobo.dir == di_f){
        set_myrobo_acc_z(-MYROBO_ACC);
    }
    else {
        set_myrobo_acc_z(MYROBO_ACC);
    }
}
void myrobo_toggle_thrust_forward(void){
    if (myrobo->myrobo.zthrust == 1){
        myrobo->myrobo.zthrust = 0;
    }
    else { // ߤޤϸλ
        myrobo->myrobo.zthrust = 1;
        myrobo_thrust_forward();
    }
}
void myrobo_toggle_thrust_backward(void){
    if (myrobo->myrobo.zthrust == -1){
        myrobo->myrobo.zthrust = 0;
    }
    else { // ߤޤλ
        myrobo->myrobo.zthrust = -1;
        myrobo_thrust_backward();
    }
}



void myrobo_halt_x(void){
    if (myrobo->myrobo.dir == di_f){
        if (myrobo->velo3.x > MYROBO_ACC){
            myrobo_thrust_left();
        }
        else if (myrobo->velo3.x < -MYROBO_ACC){
           myrobo_thrust_right();
        }
        else {
            myrobo->velo3.x = 0;
            myrobo->acc3.x = 0;
        }
    }
    else {
        if (myrobo->velo3.x < -MYROBO_ACC){
            myrobo_thrust_left();
        }
        else if (myrobo->velo3.x > MYROBO_ACC){
           myrobo_thrust_right();
        }
        else {
            myrobo->velo3.x = 0;
            myrobo->acc3.x = 0;
        }
    }
}
void myrobo_halt_y(void){
    //    printf("halt y acc:%f\n", myrobo->acc3.y);
    if (myrobo->velo3.y > MYROBO_ACC){
        myrobo_thrust_up();
    }
    else if (myrobo->velo3.y < -MYROBO_ACC){
        myrobo_thrust_down();
    }
    else {
        myrobo->velo3.y = 0;
        myrobo->acc3.y = 0;
    }
}
void myrobo_halt_z(void){
    //    printf("halt z acc:%f\n", myrobo->acc3.z);
    if (myrobo->velo3.z > MYROBO_ACC){
        if (myrobo->myrobo.dir == di_f){
            myrobo_thrust_backward();
        }
        else {
            myrobo_thrust_forward();
        }
    }
    else if (myrobo->velo3.z < -MYROBO_ACC){
        if (myrobo->myrobo.dir == di_b){
            myrobo_thrust_backward();
        }
        else {
            myrobo_thrust_forward();
        }
    }
    else {
        myrobo->velo3.z = 0;
        myrobo->acc3.z = 0;
    }
}

void myrobo_set_firing(void){
    myrobo->myrobo.is_firing = 1;
}

void myrobo_fire(void){
    obj_t *obj;
    pos3_t pos3 = myrobo->pos3;
    //  pos3.z += (myrobo->myrobo.dir == di_f) ? myrobo->size/4 : -myrobo->size/4;
    obj = create_obj(ot_mybullet, pos3);
    set_dir_mybullet(obj, myrobo->myrobo.dir);
    myrobo->myrobo.is_firing = 0;
}


void set_myrobo_acc_x(int val){
    myrobo->acc3.x = val;
};
void set_myrobo_acc_y(int val){
    myrobo->acc3.y = val;
};
void set_myrobo_acc_z(int val){
    myrobo->acc3.z = val;
};


void act_myrobo(obj_t *obj){
    move_myrobo();

    // ꥢ˽Фʤ褦ʽ
    // @todo ŪˤϤΤޤޥꥢȤƥ४Сˤ
    if (myrobo->pos3.x < 0){
        myrobo->pos3.x = 0;
        myrobo->velo3.x = 0;
    }
    else if (myrobo->pos3.x >= SPACE_XSIZE){
        myrobo->pos3.x = SPACE_XSIZE-1;
        myrobo->velo3.x = 0;
    }
    if (myrobo->pos3.y < 0){
        myrobo->pos3.y = 0;
        myrobo->velo3.y = 0;
    }
    else if (myrobo->pos3.y >= SPACE_YSIZE){
        myrobo->pos3.y = SPACE_YSIZE-1;
        myrobo->velo3.y = 0;
    }
    if (myrobo->pos3.z < 0){
        myrobo->pos3.z = 0;
        myrobo->velo3.z = 0;
    }
    else if (myrobo->pos3.z >= SPACE_ZSIZE){
        myrobo->pos3.z = SPACE_ZSIZE-1;
        myrobo->velo3.z = 0;
    }
    if (myrobo->myrobo.is_firing){
        myrobo_fire();
    }
    apply_hit_obj(obj, hit_myrobo);
}

void move_myrobo(void){
    //    printf("myrobo (%+2.2f,%+2.2f,%+2.2f)\n", myrobo->acc3.x, myrobo->acc3.y, myrobo->acc3.z);

    //    printf("turn_count:%d\n", myrobo->myrobo.turn_count);
    if (myrobo->myrobo.turn_count > 0){
        myrobo_turn_right();
    }
    else if (myrobo->myrobo.turn_count < 0){
        myrobo_turn_left();
    }

    myrobo->velo3.x += myrobo->acc3.x;
    ClipInt(myrobo->velo3.x, -MYROBO_VELO_MAX, MYROBO_VELO_MAX);
    myrobo->velo3.y += myrobo->acc3.y;
    ClipInt(myrobo->velo3.y, -MYROBO_VELO_MAX, MYROBO_VELO_MAX);

    //    printf("myrobo zthrust:%d\n", myrobo->myrobo.zthrust);
    if (myrobo->myrobo.zthrust == 0){
        myrobo_halt_z();
    }
    if (myrobo->myrobo.turn_count != 0){
        velo3_t velo_org = myrobo->velo3;
        if (myrobo->myrobo.turn_count > 0){
            myrobo->velo3.x = turn_velo[myrobo->myrobo.turn_count-1].x *
                abs(velo_org.z) / MYROBO_FVELO_MAX;
            myrobo->velo3.z = turn_velo[myrobo->myrobo.turn_count-1].y *
                abs(velo_org.z) / MYROBO_FVELO_MAX;
            if (myrobo->myrobo.turn_dir == di_f){
            }
            else {
                myrobo->velo3.x = -myrobo->velo3.x;
                myrobo->velo3.z = -myrobo->velo3.z;
            }
        }
        else {
            myrobo->velo3.x = turn_velo[-myrobo->myrobo.turn_count-1].x *
                abs(velo_org.z) / MYROBO_FVELO_MAX;
            myrobo->velo3.z = turn_velo[-myrobo->myrobo.turn_count-1].y *
                abs(velo_org.z) / MYROBO_FVELO_MAX;
            myrobo->velo3.x = -myrobo->velo3.x;
            if (myrobo->myrobo.turn_dir == di_f){
            }
            else {
                myrobo->velo3.x = -myrobo->velo3.x;
                myrobo->velo3.z = -myrobo->velo3.z;
            }
        }
        //        printf("myrobo pos:(%d, %d) velo:(%d, %d) %d\n", myrobo->pos3.x, myrobo->pos3.z, myrobo->velo3.x, myrobo->velo3.z, myrobo->myrobo.turn_count);
        move_obj(myrobo);
        myrobo->velo3 = velo_org;
    }
    else {
        myrobo->velo3.z += myrobo->acc3.z;
        if ((myrobo->velo3.z > 0 && myrobo->myrobo.dir == di_f) ||
            (myrobo->velo3.z < 0 && myrobo->myrobo.dir == di_b))
            {
                ClipInt(myrobo->velo3.z, -MYROBO_FVELO_MAX, MYROBO_FVELO_MAX);
            }
        else {
            ClipInt(myrobo->velo3.z, -MYROBO_BVELO_MAX, MYROBO_BVELO_MAX);
        }
        move_obj(myrobo);
    }
    //    printf("myrobo (%.2f, %.2f, %.2f)\n", myrobo->velo3.x, myrobo->velo3.y, myrobo->velo3.z);
}


void myrobo_start_turn_left(void){
    if (myrobo->myrobo.turn_count == 0){
        printf ("start turn left dir:%d\n", myrobo->myrobo.dir);
        myrobo->myrobo.turn_count = -1;
        myrobo->myrobo.turn_dir = myrobo->myrobo.dir;
    }
}
void myrobo_start_turn_right(void){
    if (myrobo->myrobo.turn_count == 0){
        printf ("start turn right dir:%d\n", myrobo->myrobo.dir);
        myrobo->myrobo.turn_count = 1;
        myrobo->myrobo.turn_dir = myrobo->myrobo.dir;
    }
}
void myrobo_turn_left(void){
    if (myrobo->myrobo.turn_count == -MYROBO_TURN_FRAME_NUM){
        myrobo->myrobo.turn_count = 0;
    }
    else if (myrobo->myrobo.turn_count == -MYROBO_TURN_FRAME_NUM/2-1){
        myrobo_toggle_dir();
        myrobo->myrobo.turn_count --;
    }
    else {
        myrobo->myrobo.turn_count --;
    }
}
void myrobo_turn_right(void){
    if (myrobo->myrobo.turn_count == MYROBO_TURN_FRAME_NUM){
        myrobo->myrobo.turn_count = 0;
    }
    else if (myrobo->myrobo.turn_count == MYROBO_TURN_FRAME_NUM/2+1){
        myrobo_toggle_dir();
        myrobo->myrobo.turn_count ++;
    }
    else {
        myrobo->myrobo.turn_count ++;
    }
}

void myrobo_toggle_dir(void){
    myrobo->myrobo.dir = (myrobo->myrobo.dir == di_f) ? di_b : di_f;
    printf("myrobo_dir changed to %d\n", myrobo->myrobo.dir);
    myrobo->velo3.z = -myrobo->velo3.z; 
    if (myrobo->myrobo.zthrust == 1){
        myrobo_thrust_forward();
    }
    else if (myrobo->myrobo.zthrust == -1){
        myrobo_thrust_backward();
    }
}

dir_t get_myrobo_dir(void){
    return myrobo->myrobo.dir;
}

int get_myrobo_hp(void){
    return myrobo->hp;
}

int get_myrobo_turn_count(void){
    return myrobo->myrobo.turn_count;
}
int get_myrobo_die_count(void){
    return myrobo->myrobo.die_count;
}
void inc_myrobo_die_count(void){
    myrobo->myrobo.die_count ++;
    if (myrobo->myrobo.die_count >= MYROBO_DIE_COUNT){
        set_main_state(ms_mission_select);
    }
}




double get_myrobo_velo_z(void){
    return myrobo->velo3.z;
}

void damage_myrobo(int damage){
    damage_obj(myrobo, damage);
}

int hit_myrobo(obj_t *obj, obj_t *myrobo){
    switch (obj->type){
    case ot_myrobo:
    case ot_mybullet:
    case ot_goal:
        return 0;
    default:
        if (IsExploding(obj)) return 0;
        if (is_obj_hit(obj, myrobo)){
            damage_obj(obj, BUMP_DAMAGE);
            switch (obj->type){
            case ot_dummy_target:
                break;
            case ot_mine:
                damage_obj(myrobo, MINE_DAMAGE);
                obj->hp = 0;
                break;
            // ŨƤäƤķ֤ʤ
            case ot_enemybullet:
                return 0;
            default:
                damage_obj(myrobo, BUMP_DAMAGE);
                break;
            }
            myrobo->pos3 = myrobo->prev_pos3;
            if (myrobo->velo3.x > 0){
                myrobo->velo3.x = -MYROBO_VELO_MAX;
            }
            else if (myrobo->velo3.x < 0){
                myrobo->velo3.x = MYROBO_VELO_MAX;
            }
            if (myrobo->velo3.y > 0){
                myrobo->velo3.y = -MYROBO_VELO_MAX;
            }
            else if (myrobo->velo3.y < 0){
                myrobo->velo3.y = MYROBO_VELO_MAX;
            }
            if (myrobo->velo3.z != 0){
                myrobo->myrobo.zthrust = 0;
                if (obj->rpos3.z > 0){
                    myrobo->velo3.z = -MYROBO_VELO_MAX;
                    //            obj->pos3.z += myrobo->size;
                }
                else {
                    myrobo->velo3.z = MYROBO_VELO_MAX;
                    //            myrobo->pos3.z += obj->size;
                    //            obj->pos3.z -= myrobo->size;
                }
            }
            return 1;
        }
        break;
    }
    return 0;
}

int is_myrobo_alive(void){
    return myrobo->hp;
}
