#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <EzGraph.h>
#include "config.h"
#include "param.h"
#include "obj.h"
#include "obj_draw.h"
#include "main_state.h"
#include "rnd.h"
#include "uty.h"

#define ObjName(name) ot_ ## name, #name

static obj_info_t obj_infos[] = {
    {ObjName(fragment),
     {4, {0, 0, 0},
      -1, si_invalid,
      {setup_fragment, NULL}}
    },
    {ObjName(asteroid),
     {32, {16, 16, 16},
      -1, si_asteroid,
      {setup_asteroid, NULL}}
    },
    {ObjName(mine),
     {16, {8, 8, 8},
      MINE_HP, si_mine,
      {setup_mine, NULL}}
    },
    {ObjName(myrobo),
     {24, { 8,  8,  8},
      MYROBO_HP, si_myrobo,
      {setup_myrobo, act_myrobo}}
    },
    {ObjName(drone),
     {24, { 8,  8,  8},
      DRONE_HP, si_drone_00,
      {setup_drone, act_drone}}
    },
    {ObjName(enemy_destroyer),
     {128, { 32,  8,  8},
      ENEMY_DESTROYER_HP, si_enemy_destroyer,
      {setup_enemy_destroyer, act_enemy_destroyer}}
    },
    {ObjName(dummy_target),
     {24, { 8,  8,  8},
      DUMMY_TARGET_HP, si_dummy_target,
      {setup_dummy_target, NULL}}
    },
    {ObjName(goal),
     {64, {64, 64, 64},
      -1, si_invalid,
      {setup_goal, act_goal}}
    },
    {ObjName(mybullet),
     {2, {2, 2, 8},
      -1, si_invalid,
      {setup_mybullet, act_mybullet}}
    },
    {ObjName(enemybullet),
     {2, {2, 2, 8},
      -1, si_invalid,
      {setup_enemybullet, act_enemybullet}}
    },
};
#define OBJ_INFOS_NUM (sizeof(obj_infos)/sizeof(obj_info_t))

static obj_table_entry_t obj_table[ot_end];
//obj_callbacks_t obj_callback_table[ot_end];

static obj_t objs[OBJ_NUM_MAX]; // Ȥꤢǳݡ
// ̤ѥ֥ݻ
// push/popʤñꥹ
static obj_t *free_obj_list = NULL;
// (¸Ƥ)֥ݻ
// ֥ȤΥʤɤϤΥꥹȤʤƼ¹Ԥ
// Ȥ˥ꥹ椫Τꥹ
static obj_t *alive_obj_list = NULL;

static int zbuf_ixs[SPACE_ZSIZE];
static obj_t *zbuf[SPACE_ZSIZE][ZBUF_OBJ_MAX];

// ƥ֥ȤȽzκ
// zbufȽåϰϤιʤߤ˻
static int hitbox_z_max = 0;


int obj_init(void){
    int i;

    
    
    for (i = 0 ; i < OBJ_INFOS_NUM ; i++){
        obj_table[obj_infos[i].type] = obj_infos[i].entry;
        if (hitbox_z_max < obj_table[obj_infos[i].type].hitbox.z){
            hitbox_z_max = obj_table[obj_infos[i].type].hitbox.z;
        }
    }
    obj_reset();
#ifdef DEBUG
    print_obj_list(free_obj_list);
#endif
    myrobo_init();
    drone_init();

    return 0;
}

obj_type_t obj_name2type(char *str){
    int i;
    for (i = 0 ; i < OBJ_INFOS_NUM; i++){
        if (strcmp(str, obj_infos[i].name) == 0){
            return obj_infos[i].type;
        }
    }
    return ot_invalid;
}

void obj_reset(void){
    int i;

    alive_obj_list = NULL;
    free_obj_list = objs;

    for (i = 0 ; i < OBJ_NUM_MAX-1 ; i++){
        objs[i].next = &objs[i+1];
    }
    objs[OBJ_NUM_MAX-1].next = NULL;
#ifdef DEBUG
    for (i = 0 ; i < OBJ_NUM_MAX ; i++){
        objs[i].id = i;
    }
#endif    
}

void push_obj_uni(obj_t **list, obj_t *obj){
    obj->next = *list;
    *list = obj;
}
obj_t *pop_obj_uni(obj_t **list){
    obj_t *new_obj;

    if (*list == NULL) return NULL;
    new_obj = *list;
    *list = (*list)->next;
    return new_obj;
}
void push_obj_bi(obj_t **list, obj_t *obj){
    obj->prev = NULL;
    obj->next = *list;
    if (*list != NULL) {
        (*list)->prev = obj;
    }
    *list = obj;
}
obj_t *pop_obj_bi(obj_t **list){
    obj_t *new_obj;

    if (*list == NULL) return NULL;
    new_obj = *list;
    *list = (*list)->next;
    (*list)->prev = NULL;
    return new_obj;
}
void delete_obj_bi(obj_t **list, obj_t *obj){
    assert(*list != NULL);
    assert(obj != NULL);

    // ƬǤκ
    if (*list == obj){
        *list = obj->next;
        if (obj->next != NULL){
            obj->next->prev = NULL;
        }
    }
    else {
        assert(obj->prev != NULL); // objƬǤʤǤ¸ߤϤ
        obj->prev->next = obj->next;
        if (obj->next != NULL){
            obj->next->prev = obj->prev;
        }
    }
}


// ֥Ȥ
// ե꡼ꥹȤļФ֤
// κݤˡ¸֥ȥꥹȤˤ⼫ưϿ
obj_t *alloc_obj(void){
    obj_t *new_obj;

    new_obj = pop_obj_uni(&free_obj_list);
    // ꥹȤʤȤꤢ
    // @note ̿ŪĤalloc_objΤǡ
    // ƤӽФ¦˥顼Ǥˤǻ
    assert(new_obj != NULL);

    push_obj_bi(&alive_obj_list, new_obj);
    return new_obj;
}

void free_obj(obj_t *obj){
    delete_obj_bi(&alive_obj_list, obj);
    push_obj_uni(&free_obj_list, obj);
}

void for_each_obj(obj_callback_func_t func){
    obj_t *obj = alive_obj_list;
    obj_t *obj_next;
    
    while (obj != NULL){
        // funcobjǽ뤿ᡢnext򤿤ɤäƤ
        obj_next = obj->next;
        func(obj);
        obj = obj_next;
    }
}


// func0֤ϡǽǤڤ0֤
// Ǹޤǽ줿Ȥ0֤
// base_objNULLǤʤȤϡobj==base_objΤȤϥХåʤ
int for_each_obj_zbuf(obj_callback_func_t func, int zst, int zen){
    int z, i;
    obj_t *obj;

    assert(zst != zen);

    if (zst > zen){
        if (zst >= SPACE_ZSIZE) zst = SPACE_ZSIZE-1;
        if (zen < 0) zen = 0;
        for (z = zst ; z >= zen ; z--){
            for (i = 0 ; i < zbuf_ixs[z] ; i++){
                obj = zbuf[z][i];
                if (func(obj)) return 1;
            }
        }
    }
    else {
        if (zen >= SPACE_ZSIZE) zen = SPACE_ZSIZE-1;
        if (zst < 0) zst = 0;
        for (z = zst ; z <= zen ; z++){
            for (i = 0 ; i < zbuf_ixs[z] ; i++){
                obj = zbuf[z][i];
                if (func(obj)) return 1;
            }
        }
    }
    return 0;
}
int for_each_obj_zbuf2(obj_callback_func2_t func, int zst, int zen,
                       obj_t *base_obj)
{
    int z, i;
    obj_t *obj;

    assert(zst != zen);

    if (zst > zen){
        if (zst >= SPACE_ZSIZE) zst = SPACE_ZSIZE-1;
        if (zen < 0) zen = 0;
        for (z = zst ; z >= zen ; z--){
            for (i = 0 ; i < zbuf_ixs[z] ; i++){
                obj = zbuf[z][i];
                if (func(obj, base_obj)) return 1;
            }
        }
    }
    else {
        if (zen >= SPACE_ZSIZE) zen = SPACE_ZSIZE-1;
        if (zst < 0) zst = 0;
        for (z = zst ; z <= zen ; z++){
            for (i = 0 ; i < zbuf_ixs[z] ; i++){
                obj = zbuf[z][i];
                if (func(obj, base_obj)) return 1;
            }
        }
    }
    return 0;
}

void reset_obj_zbuf(void){
    memset(zbuf_ixs, 0, sizeof(int)*SPACE_ZSIZE);
}

void register_obj_zbuf(obj_t *obj){
    int z = (int)obj->pos3.z;

    // ƱZɸǾĶ顢ȤꤢϿʤ
    // @note ɽϷ礱ȽʤɤǤ̵뤵Τ
    // ºݤ˵ȤޤZBUF_OBJ_MAXʤ礭Фޤʤ
    if (zbuf_ixs[z] >= ZBUF_OBJ_MAX) return;

    // ֳΥ֥ȤϿʤ
    // @todo Ʈ礭SPACE礭ꤷ
    // ϤޤǹԤʤ֥Ȥ¸ߤɽ줦ΰ
    if (z >= SPACE_ZSIZE || z < 0) return;

    zbuf[z][zbuf_ixs[z]] = obj;
    zbuf_ixs[z]++;
}


obj_t *create_obj(obj_type_t type, pos3_t pos3){
    obj_t *new_obj;
    new_obj = alloc_obj();
    assert(new_obj != NULL);

    new_obj->type = type;
    new_obj->sid = obj_table[type].sid;
    new_obj->size = obj_table[type].size;
    new_obj->hp = obj_table[type].hp_max;
    new_obj->valid = 1;
    new_obj->explosion_count = 0;
    new_obj->pos3 = pos3;
    if (obj_table[type].callbacks.setup != NULL){
        obj_table[type].callbacks.setup(new_obj);
    }
    return new_obj;
}

obj_t *create_obj_region(obj_type_t type, region3_t region3, int num){
    int i;
    pos3_t pos3;
    obj_t *obj;

    for (i = 0 ; i < num ; i++){
        pos3.x = RRnd(region3.st.x, region3.en.x);
        pos3.y = RRnd(region3.st.y, region3.en.y);
        pos3.z = RRnd(region3.st.z, region3.en.z);
        obj = create_obj(type, pos3);
        if (obj == NULL) return NULL;
    }
    return obj; // NULLåѤ˺Ǹobj֤Ƥ
}

void delete_invalid_objs(void){
    for_each_obj(delete_invalid_obj);
}
int delete_invalid_obj(obj_t *obj){
    if (obj->valid) return 0;
    // ȤꤢŪ˼˲Υ४Сľܽ񤤤Ƥ
    if (obj->type == ot_myrobo){
        inc_myrobo_die_count();
        // myroborpos׻δʤΤǡfreeƤϤʤ
    }
    else {
        free_obj(obj);
    }
    return 0;
}

void act_objs(void){
    for_each_obj(act_obj);
}

int act_obj(obj_t *obj){
    //    printf("act obj type:%d\n", obj->type);
    if (obj->hp == 0){
        obj->explosion_count ++;
        //        printf("ex count:%d\n", obj->explosion_count);
        // ȯä餽ǤΥ֥Ȥ̵
        // ºݤνdelete_objsǤޤȤƤ
        if (obj->explosion_count >= OBJ_EXPLOSION_COUNT_MAX){
    // @todo Ūˤϡǥ֥Ȥdie٥ȥեå
            obj->valid = 0;
        }
        return 0;
    }
    if (obj_table[obj->type].callbacks.act != NULL){
        obj_table[obj->type].callbacks.act(obj);
    }
    return 0;
}

void update_objs(void){
    pos3_t base_pos3;
    obj_t *obj;

    // ZХåեϿ
    obj = alive_obj_list;
    reset_obj_zbuf();
    while (obj != NULL){
        register_obj_zbuf(obj);
        obj = obj->next;
    }
    
    // кɸ򹹿
    obj = alive_obj_list;
    get_myrobo_pos3(&base_pos3);
    while (obj != NULL){
        obj->rpos3.x = obj->pos3.x - base_pos3.x;
        obj->rpos3.y = obj->pos3.y - base_pos3.y;
        obj->rpos3.z = obj->pos3.z - base_pos3.z;
        obj = obj->next;
    }
}


int get_obj_size(obj_type_t type){
    return obj_table[type].size;
}

void move_obj(obj_t *obj){
    obj->prev_pos3 = obj->pos3;
    obj->pos3.x += obj->velo3.x;
    obj->pos3.y += obj->velo3.y;
    obj->pos3.z += obj->velo3.z;
}

#define HitDistX(obj1, obj2)                    \
    (obj_table[(obj1)->type].hitbox.x +         \
     obj_table[(obj2)->type].hitbox.x)
#define HitDistY(obj1, obj2)                    \
    (obj_table[(obj1)->type].hitbox.y +         \
     obj_table[(obj2)->type].hitbox.y)
#define HitDistZ(obj1, obj2)                    \
    (obj_table[(obj1)->type].hitbox.z +         \
     obj_table[(obj2)->type].hitbox.z)
    

int is_obj_hit(obj_t *obj1, obj_t *obj2){
    if (obj1->explosion_count != 0 || obj2->explosion_count != 0) return 0;
    //  printf("check hit (%d, %d, %d) (%d, %d, %d)\n", obj1->pos3.x, obj1->pos3.y, obj1->pos3.z, obj2->pos3.x, obj2->pos3.y, obj2->pos3.z);
    if (abs(obj1->pos3.x-obj2->pos3.x) <= HitDistX(obj1, obj2) &&
        abs(obj1->pos3.y-obj2->pos3.y) <= HitDistY(obj1, obj2) &&
        abs(obj1->pos3.z-obj2->pos3.z) <= HitDistZ(obj1, obj2))
    {
        puts("hit!");
        return 1;
    }
    return 0;
}

int is_obj_hit_myrobo(obj_t *obj){
    if (abs(obj->rpos3.x) <= obj_table[(obj)->type].hitbox.x &&
        abs(obj->rpos3.y) <= obj_table[(obj)->type].hitbox.y &&
        abs(obj->rpos3.z) <= obj_table[(obj)->type].hitbox.z)
    {
        puts("myrobo hit!");
        return 1;
    }
    return 0;
}


void apply_hit_obj(obj_t *obj, obj_callback_func2_t func){
    for_each_obj_zbuf2(func,
                       (int)obj->pos3.z - hitbox_z_max,
                       (int)obj->pos3.z + hitbox_z_max,
                       obj);
}

void damage_obj(obj_t *obj, int damage){
    // HPΥ֥Ȥ˲ǽ
    if (obj->hp < 0) return;
    obj->hp -= damage;
    if (obj->hp <= 0){
        obj->hp = 0;
    }
    printf("damage %d HP:%d\n", damage, obj->hp);
}



#ifdef DEBUG
void print_obj_list(obj_t *list){
    while (list != NULL){
        print_obj(list);
        list = list->next;
    }
}
void print_obj(obj_t *obj){
    //    printf("[%3d]%d (%d, %d, %d)\n", obj->id, obj->type, obj->pos3.x, obj->pos3.y, obj->pos3.z);
}

void print_free_list(void){
    puts("free list:");
    print_obj_list(free_obj_list);
}

void print_alive_list(void){
    puts("alive list:");
    print_obj_list(alive_obj_list);
}


#endif
