#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <EzGraph.h>
#include "config.h"
#include "coordinate.h"
#include "param.h"
#include "layout.h"
#include "obj.h"
#include "obj_draw.h"
#include "sprite.h"

void draw_default(int x, int y, int size, obj_t *obj);
//void draw_asteroid(int x, int y, int size, obj_t *obj);
//void draw_mine(int x, int y, int size, obj_t *obj);
void draw_myrobo(int x, int y, int size, obj_t *obj);
void draw_enemy_destroyer(int x, int y, int size, obj_t *obj);
void draw_drone(int x, int y, int size, obj_t *obj);
void draw_mybullet(int x, int y, int size, obj_t *obj);
void draw_enemybullet(int x, int y, int size, obj_t *obj);
void draw_fragment(int x, int y, int size, obj_t *obj);
void draw_goal(int x, int y, int size, obj_t *obj);

typedef void(*obj_draw_callback_t)(int x, int y, int size, obj_t *obj);

static dir_t screen_dir;

typedef struct {
    obj_type_t type;
    //    void(*draw)(int x, int y, int size, obj_t *obj);
    obj_draw_callback_t draw;
} obj_draw_info_t;

static obj_draw_info_t obj_draw_infos[] = {
    {ot_asteroid, draw_default},
    {ot_mine, draw_default},
    {ot_myrobo, draw_myrobo},
    {ot_enemy_destroyer, draw_enemy_destroyer},
    {ot_dummy_target, draw_default},
    {ot_drone, draw_drone},
    {ot_mybullet, draw_mybullet},
    {ot_enemybullet, draw_enemybullet},
    {ot_fragment, draw_fragment},
    {ot_goal, draw_goal},
};
#define OBJ_DRAW_INFOS_NUM (sizeof(obj_draw_infos)/sizeof(obj_draw_info_t))

obj_draw_callback_t obj_draw_callback_table[ot_end];

sprite_id_t sid_myrobo_turn_l[] = {
    si_myrobo_turn_l1,
    si_myrobo_turn_l2,
    si_myrobo_turn_l3,
    si_myrobo_turn_l4,
    si_myrobo_turn_l5,
    si_myrobo_turn_l6,
    si_myrobo_turn_l7,
    si_myrobo_turn_l8,
    si_myrobo_turn_l9,
    si_myrobo_turn_r8,
    si_myrobo_turn_r7,
    si_myrobo_turn_r6,
    si_myrobo_turn_r5,
    si_myrobo_turn_r4,
    si_myrobo_turn_r3,
    si_myrobo_turn_r2,
    si_myrobo_turn_r1,
};
sprite_id_t sid_myrobo_turn_r[] = {
    si_myrobo_turn_r1,
    si_myrobo_turn_r2,
    si_myrobo_turn_r3,
    si_myrobo_turn_r4,
    si_myrobo_turn_r5,
    si_myrobo_turn_r6,
    si_myrobo_turn_r7,
    si_myrobo_turn_r8,
    si_myrobo_turn_r9,
    si_myrobo_turn_l8,
    si_myrobo_turn_l7,
    si_myrobo_turn_l6,
    si_myrobo_turn_l5,
    si_myrobo_turn_l4,
    si_myrobo_turn_l3,
    si_myrobo_turn_l2,
    si_myrobo_turn_l1,
};

void obj_draw_init(void){
    int i;

    for (i = 0 ; i < OBJ_DRAW_INFOS_NUM ; i++){
        obj_draw_callback_table[obj_draw_infos[i].type]
            = obj_draw_infos[i].draw;
    }
    screen_dir = di_f;
}

void draw_screen_objs(void){ 
    pos3_t base_pos3;
    int z;
    get_myrobo_pos3(&base_pos3);
    z = (int)base_pos3.z;

    screen_dir = get_myrobo_dir();
    if (screen_dir == di_f){
        for_each_obj_zbuf(draw_screen_obj, z+SCREEN_RZMAX, z+SCREEN_RZMIN);
    }
    else {
        for_each_obj_zbuf(draw_screen_obj, z-SCREEN_RZMAX, z-SCREEN_RZMIN);
    }
} 



void draw_default(int x, int y, int size, obj_t *obj){
    //    if (obj->type == ot_enemy_destroyer){
        //        printf("destroyer size=%d\n", size);
    //    }
    if (obj->hp == 0){
        draw_sprite(x, y, size, screen_dir, si_explosion);
    }
    else {
        draw_sprite(x, y, size, screen_dir, obj->sid);
    }
}
void draw_myrobo(int x, int y, int size, obj_t *obj){
    int ctr;

    if (get_myrobo_die_count() > 0) return;

    ctr = get_myrobo_turn_count();
    //    printf("turn count:%d\n", ctr);
    if (ctr > 0){
        draw_sprite(x, y, size, screen_dir, sid_myrobo_turn_r[ctr-2]);
    }
    else if (ctr < 0){
        draw_sprite(x, y, size, screen_dir, sid_myrobo_turn_l[-ctr-2]);
    }
    else {
        draw_default(x, y, size, obj);
    }
    // !!! ʮͱʤɤ
}
void draw_enemy_destroyer(int x, int y, int size, obj_t *obj){
    if (obj->hp == 0){
        draw_sprite(x-size/3, y-size/24, size/5, screen_dir, si_explosion);
        draw_sprite(x-size/6, y+size/16,  size/6, screen_dir, si_explosion);
        draw_sprite(x,        y,         size/4, screen_dir, si_explosion);
        draw_sprite(x+size/4, y,         size/4, screen_dir, si_explosion);
        draw_sprite(x+size/3, y+size/16, size/6, screen_dir, si_explosion);
    }
    else {
        draw_default(x, y, size, obj);
    }
}
void draw_drone(int x, int y, int size, obj_t *obj){
    draw_default(x, y, size, obj);
}
    
void draw_mybullet(int x, int y, int size, obj_t *obj){
    EzSetColorByRGB(255, 255, 0);
//    EzFillBoxB(x-size/2, y-size/2, x+size/2, y+size/2);
//    EzFillCircleB(x, y, size/2);
    EzFillCircleB(x, y, size);
}
void draw_enemybullet(int x, int y, int size, obj_t *obj){
    // @todo Ū˥᡼̤ƤμȽ
    // Ūˤϼ̥եɤɲä롣
    if (obj->enemybullet.damage == DESTROYER_BULLET_DAMAGE){
        EzSetColorByRGB(255, 255, 0);
    }
    else {
        EzSetColorByRGB(255, 0, 0);
    }
//    EzFillBoxB(x-size/2, y-size/2, x+size/2, y+size/2);
    EzFillCircleB(x, y, size/2);
}
#define FragmentSize(size) ((size)/24+1)
void draw_fragment(int x, int y, int size, obj_t *obj){
    //    EzSetColorByRGB(120, 120, 120);
    // EzSetColorByRGB(255, 255, 255);
    EzSetColorByRGB(160, 160, 160);
//    EzFillBoxB(x-size/2, y-size/2, x+size/2, y+size/2);
//    if (size > 8 && abs(obj->rpos3.x) < 128 && abs(obj->rpos3.y)<128) return;
    if (abs(obj->rpos3.x) < 128 && abs(obj->rpos3.y)<128) return;
        //        EzFillCircleB(x, y, size/4);
    EzFillBoxB(x-FragmentSize(size), y-FragmentSize(size),
               x+FragmentSize(size), y+FragmentSize(size));
//    EzFillCircleB(x, y, 50);
}
void draw_goal(int x, int y, int size, obj_t *obj){
    EzSetColorByRGB(200, 200, 200);
//    EzFillBoxB(x-size/2, y-size/2, x+size/2, y+size/2);
    EzDrawCircleB(x, y, size);
    EzDrawCircleB(x, y, size*3/4);
}

int draw_screen_obj(obj_t *obj){
    pos3_t base_pos3;
    int size, x, y;
    double rx, ry, rz;

    get_myrobo_pos3(&base_pos3);
    rx = obj->rpos3.x;
    ry = obj->rpos3.y;
    rz = obj->rpos3.z;

    if (screen_dir == di_f){
        size = Size2ScreenSize(obj->size, rz);
    //   printf("rz:%f size:%d sprite_size:%d\n", rz, obj->size, size);
        x = Rx2ScreenX(rx, rz);
        y = Ry2ScreenY(ry, rz);
    }
    else {
        size = Size2ScreenSize(obj->size, -rz);
    //   printf("rz:%f size:%d sprite_size:%d\n", rz, obj->size, size);
        x = Rx2ScreenX(-rx, -rz);
        y = Ry2ScreenY(ry, -rz);
    }
    //            size = 64/(1+rz/SPACE_ZSIZE*32);
        //        x = 0;
    //        x = SCREEN_W * rx / (SCREEN_DEPTH+rz);
    //        y = SCREEN_W * ry / (SCREEN_DEPTH+rz);
        //        x = 32 * rx / abs(rz);
        //        y = 32 * ry / abs(rz);
        //    }
    //        if (abs(unit_pos.x-myrobo_pos) < 

    //    printf("%p\n", obj_draw_callback_table[obj->type]);
    assert(obj_draw_callback_table[obj->type]);
    obj_draw_callback_table[obj->type](x, y, size, obj);
    EzSetColorByRGB(255, 0, 0);
    //    EzDrawBoxB(x-1, y-1, x+1, y+1);
    return 0;
}

void draw_map_objs(void){ 
    pos3_t pos3;
    int ltx, lty, rbx, rby;

    get_myrobo_pos3(&pos3);
    EzSetColorByRGB(255, 255, 0);
    ltx = X2MapX(pos3.x-RADAR_RANGE);
    lty = Z2MapY(pos3.z+RADAR_RANGE);
    rbx = X2MapX(pos3.x+RADAR_RANGE);
    rby = Z2MapY(pos3.z-RADAR_RANGE);
    //    EzDrawCircleB(x, y, MAP_XSIZE*RADAR_RANGE/SPACE_XSIZE);
    //    EzDrawCircleB(x, y, Size2MapSize(RADAR_RANGE));
    EzDrawBoxB(ltx, lty, rbx, rby);

    for_each_obj(draw_map_obj);
} 
int draw_map_obj(obj_t *obj){
    int x, y;

    switch (obj->type){
    case ot_myrobo:
        SetMyroboColor();
        break;
    case ot_dummy_target:
    case ot_drone:
    case ot_enemy_destroyer:
        SetEnemyColor();
        break;
    case ot_asteroid:
        SetNeutralTerranColor();
        break;
    case ot_mine:
        SetHostileTerranColor();
        break;
    case ot_goal:
        SetGoalColor();
        break;
    case ot_fragment:
        //        EzSetColorByRGB(255, 0, 0);
        //        break;
    // ޥåפˤɽʤ֥
    case ot_mybullet:
    case ot_enemybullet:
        return 0;
    default:
        printf("unknown object %d\n", obj->type);
        assert(0);
        break;
    }
    
    x = X2MapX(obj->pos3.x);
    y = Z2MapY(obj->pos3.z);

    EzFillBoxB(x-1, y-1, x+1, y+1);
    //    printf("%d map(%d, %d)\n", obj->id, x, z);
    return 0;
}

#ifdef DEBUG
void draw_wsensor_objs(void){
    for_each_obj(draw_wsensor_obj);
}

int draw_wsensor_obj(obj_t *obj){
    int x, y;

    switch (obj->type){
    case ot_myrobo:
        EzSetColorByRGB(160, 160, 0);
        break;
    case ot_asteroid:
        EzSetColorByRGB(160, 160, 160);
        break;
    case ot_mine:
        EzSetColorByRGB(0, 160, 0);
        break;
    case ot_drone:
        EzSetColorByRGB(160, 0, 0);
        break;
    // wsensorˤɽʤ֥
    case ot_goal:
    case ot_mybullet:
    case ot_enemybullet:
    case ot_fragment:
        return 0;
    default:
        assert(0);
        break;
    }
    x = (int)(obj->pos3.x * WSENSOR_XSIZE / SPACE_XSIZE + WSENSOR_LTX);
    y = (int)(obj->pos3.y * WSENSOR_YSIZE / SPACE_YSIZE + WSENSOR_LTY);
    EzFillBoxB(x-1, y-1, x+1, y+1);
    EzFillBoxB(x-2, y-2, x+2, y+2);
    //    printf("%d map(%d, %d)\n", obj->id, x, y);
    return 0;
}
#endif

void draw_radar_objs(void){ 
    pos3_t base_pos3;
    int z;

    get_myrobo_pos3(&base_pos3);
    z = (int)base_pos3.z;

    if (screen_dir == di_f){
        for_each_obj_zbuf(draw_radar_f_obj, z+RADAR_RANGE, z);
        for_each_obj_zbuf(draw_radar_b_obj, z-RADAR_RANGE, z);
    }
    else {
        for_each_obj_zbuf(draw_radar_f_obj, z-RADAR_RANGE, z);
        for_each_obj_zbuf(draw_radar_b_obj, z+RADAR_RANGE, z);
    }
}
int draw_radar_f_obj(obj_t *obj){
    draw_radar_obj(obj, RADAR_F_X, RADAR_F_Y);
    return 0;
}
int draw_radar_b_obj(obj_t *obj){
    draw_radar_obj(obj, RADAR_B_X, RADAR_B_Y);
    return 0;
}

void draw_radar_obj(obj_t *obj, int rx, int ry){
	double at_yz, at_xz, theta, dist2;
    // double dist;
    int br;

	dist2 = Vec3Size2(obj->rpos3);
    if (dist2 > RADAR_RANGE2) return;
    br = 255 - (int)(dist2 * 100 / (RADAR_RANGE2));
    switch (obj->type){
    case ot_asteroid:
        EzSetColorByRGB(br, br, br);
        break;
    case ot_dummy_target:
    case ot_enemy_destroyer:
    case ot_drone:
        EzSetColorByRGB(br, 0, 0);
        break;
    case ot_mine:
        EzSetColorByRGB(br, 0, br);
        break;
    case ot_goal:
        //        EzSetColorByRGB(br, br, br);
        EzSetColorByRGB(255, 255, 255);
        break;
    // 졼ˤɽʤ֥
    case ot_myrobo:
    case ot_mybullet:
    case ot_enemybullet:
    case ot_fragment:
        return;
    default:
        assert(0);
        break;
    }

    // !!! dist?
    //	dist = sqrt(dist2);

    if (obj->rpos3.z == 0){
        at_yz = at_xz = 0;
    }
    else {
        at_yz = atan2(obj->rpos3.y, obj->rpos3.z);
        at_xz = atan2(obj->rpos3.x, obj->rpos3.z);
    }

    if (obj->rpos3.z < 0){
        if (at_xz > 0){
            at_xz = PI - at_xz;
        }
        else {
            at_xz = -PI - at_xz;
        }
        if (at_yz > 0){
            at_yz = PI - at_yz;
        }
        else {
            at_yz = -PI - at_yz;
        }
    }        
    
    /*    
    if (obj->rpos3.z >= 0){
        x = RADAR_F_X;
        y = RADAR_F_Y;
    }
    else {
        if (at_xz > 0){
            at_xz = PI - at_xz;
        }
        else {
            at_xz = -PI - at_xz;
        }
        if (at_yz > 0){
            at_yz = PI - at_yz;
        }
        else {
            at_yz = -PI - at_yz;
        }
        x = RADAR_B_X;
        y = RADAR_B_Y;
    }
    */  
    if (at_xz == 0){
        theta = 0;
    }
    else {
        theta = atan2(at_yz, at_xz);
    }
    if (screen_dir == di_f){
        rx += (int)(cos(theta)*RADAR_R*fabs(at_xz) / (PI/2));
    }
    else {
        rx -= (int)(cos(theta)*RADAR_R*fabs(at_xz) / (PI/2));
    }
    ry += (int)(sin(theta)*RADAR_R*fabs(at_yz) / (PI/2));
    
#ifdef DUMMY    
    if (obj->rpos3.z >= 0){
        /*
        if (IsDelta(obj->rpos3.z)){
            theta = atan2(at_yz, at_xz);
        }
        else {
            theta = atan2(at_yz, at_xz);
        }
        */
        if (at_xz == 0){
            theta = 0;
        }
        else {
            theta = atan2(at_yz, at_xz);
        }
        x = cos(theta)*RADAR_R*fabs(at_xz) / (PI/2);
        y = sin(theta)*RADAR_R*fabs(at_yz) / (PI/2);
        if (screen_dir == di_b){
            x = -x;
        }
        x += RADAR_F_X;
        y += RADAR_F_Y;
    }
    else {
        if (at_xz > 0){
            at_xz = PI - at_xz;
        }
        else {
            at_xz = -PI - at_xz;
        }
        if (at_yz > 0){
            at_yz = PI - at_yz;
        }
        else {
            at_yz = -PI - at_yz;
        }

        theta = atan2(at_yz, at_xz);
        x = cos(theta)*RADAR_R*fabs(at_xz) / (PI/2);
        y = sin(theta)*RADAR_R*fabs(at_yz) / (PI/2);
        if (screen_dir == di_b){
            x = -x;
        }
        //        printf("B (%.2f, %.2f, %.2f) (%.2f, %.2f) (%d, %d)\n", obj->rpos3.x, obj->rpos3.y, obj->rpos3.z, at_xz, at_yz, x, y);
        x += RADAR_B_X;
        y += RADAR_B_Y;
#endif

    //    EzFillBoxB(rx-1, ry-1, rx+1, ry+1);
    //    EzFillBoxB(rx-2, ry-2, rx+2, ry+2);
    EzFillBoxB(rx-1, ry-1, rx+1, ry+1);
    //    printf("%d map r(%.2f, %.2f, %.2f) 2d(%d, %d) yz:%2.2f xz:%2.2f\n", obj->id, obj->rpos3.x, obj->rpos3.y, obj->rpos3.z, x, y, at_yz*180/3.14, at_xz*180/3.14);
}

