#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>
#include "process_xpm.h"

#define PixelStr2Code(s) (s.chr[0]+(s.chr[1]<<7))
#define PixelStr(s, ix) (s.chr[ix])

#define DXpmPixel(dxpm, x, y) ((dxpm)->pixels[(y)*((dxpm)->width)+(x)])
#define IsSamePixelStr(s1, s2) \
    (PixelStr(s1, 0)==PixelStr(s2, 0) &&        \
     PixelStr(s1, 1)==PixelStr(s2, 1))
        


int is_color_name_transparent(char *name){
    char *p;

    p = name;
    while (*p != '\0'){
        *p = toupper(*p);
        p++;
    }
    if (strcmp(name, "NONE") == 0) return 1;
    return 0;
}

void copy_dxpm_color(dynamic_xpm_t *src_dxpm, dynamic_xpm_t *dest_dxpm){
    int i;

    dest_dxpm->color_num = src_dxpm->color_num;
    dest_dxpm->depth = src_dxpm->depth;
    dest_dxpm->transparent_str = src_dxpm->transparent_str;
    for (i = 0 ; i < src_dxpm->color_num ; i++){
        dest_dxpm->colors[i] = src_dxpm->colors[i];
    }
}


dynamic_xpm_t *alloc_dynamic_xpm(int width, int height){
    dynamic_xpm_t *dxpm;

    dxpm = malloc(sizeof(dynamic_xpm_t));
    if (dxpm == NULL) return NULL;

    dxpm->width = dxpm->width_max = width;
    dxpm->height = dxpm->height_max = height;

    dxpm->pixels = malloc(sizeof(pixel_str_t)*width*height);
    printf("dxpmp->pixels:%p (%dx%d)\n", dxpm->pixels, width, height);
    if (dxpm->pixels == NULL){
        free(dxpm);
        return NULL;
    }

    return dxpm;
}

void free_dynamic_xpm(dynamic_xpm_t *dxpm){
    free(dxpm->pixels);
    free(dxpm);
}

char **alloc_xpm_buffer(int width, int height, int color_num){
    char **xpm;
    char *buf;
    int line_width; // xpmʸκĹ
    int line_num; // xpmʸιԿ
    int i;

    line_width = (width*DEPTH_MAX > XPM_COLOR_NAME_LEN_MAX) ?
        width*DEPTH_MAX : XPM_COLOR_NAME_LEN_MAX;
    line_num = 1 + color_num + height;

    xpm = malloc(sizeof(char*) * line_num);
    if (xpm == NULL) return NULL;
    buf = malloc((line_width+1) * line_num);
    if (buf == NULL) return NULL;  

    for (i = 0 ; i < line_num ; i++){
        xpm[i] = buf;
        buf += line_width+DEPTH_MAX + 1;
    }
    return xpm;
}

void free_xpm_buffer(char **xpm){
    free(xpm[0]); // bufβ
    free(xpm);
}

int xpm2dynamic_xpm(char **xpm, dynamic_xpm_t *dxpm){
    int i, ix, x, y, xix;
    int width, height, color_num, depth;
    pixel_str_t pixel_str;

    if (sscanf(xpm[0], "%d %d %d %d", &width, &height, &color_num, &depth)
        != 4)
    {
        return 1;
    }
    dxpm->width = width;
    dxpm->height = height;
    dxpm->color_num = color_num;
    dxpm->depth = depth;
    
    PixelStr(dxpm->transparent_str, 0) = 0;
    PixelStr(dxpm->transparent_str, 1) = 0;
    ix = 1;
    for (i = 0 ; i < color_num ; i++){
        PixelStr(dxpm->colors[i].str, 0) = xpm[ix][0];
        if (depth == 1){
            PixelStr(dxpm->colors[i].str, 1) = '\0';
        }
        else {
            PixelStr(dxpm->colors[i].str, 1) = xpm[ix][1];
        }
        xix = depth;
        while (xpm[ix][xix] != 'c') xix++;
        xix++;
        while (xpm[ix][xix] == ' ') xix++;
        strcpy(dxpm->colors[i].name, &xpm[ix][xix]);
        if (is_color_name_transparent(dxpm->colors[i].name)){
            dxpm->transparent_str = dxpm->colors[i].str;
        }
        ix ++;
    }
    for (y = 0 ; y < height ; y++){
        for (x = 0 ; x < width ; x++){
            if (depth == 1){
                PixelStr(pixel_str, 0) = xpm[ix][x];
                PixelStr(pixel_str, 1) = '\0';
            }
            else {
                PixelStr(pixel_str, 0) = xpm[ix][x*2];
                PixelStr(pixel_str, 1) = xpm[ix][x*2+1];
            }
            DXpmPixel(dxpm, x, y) = pixel_str;
        }
        ix ++;
    }
    return 0;
}

int dynamic_xpm2xpm(dynamic_xpm_t *dxpm, char **xpm){
    int ix, i, x, y;
    sprintf(xpm[0], "%d %d %d %d",
            dxpm->width, dxpm->height, dxpm->color_num, dxpm->depth);
    ix = 1;
    for (i = 0 ; i < dxpm->color_num ; i++){
        if (dxpm->depth == 1){
            sprintf(xpm[ix], "%c c %s",
                    PixelStr(dxpm->colors[i].str, 0), dxpm->colors[i].name);
        }
        else {
            sprintf(xpm[ix], "%c%c c %s",
                    PixelStr(dxpm->colors[i].str, 0),
                    PixelStr(dxpm->colors[i].str, 1),
                    dxpm->colors[i].name);
        }
        ix ++;
    }
    for (y = 0 ; y < dxpm->height ; y++){
        if (dxpm->depth == 1){
            for (x = 0 ; x < dxpm->width ; x++){
                xpm[ix][x] = PixelStr(DXpmPixel(dxpm, x, y), 0);
            }
            xpm[ix][dxpm->width] = '\0';
        }
        else {
            for (x = 0 ; x < dxpm->width ; x++){
                xpm[ix][x*2] = PixelStr(DXpmPixel(dxpm, x, y), 0);
                xpm[ix][x*2+1] = PixelStr(DXpmPixel(dxpm, x, y), 1);
            }
            xpm[ix][dxpm->width*2] = '\0';
        }
        ix ++;
    }
    return 0;
}

// lower_dxpmξupper_dxpmŤͤ
// !!! ߤ»
// - lowerȤupperƱ(lowerΥ֥å)ǤʤФʤʤ
// - 礭lowerƱˤʤ롣
//   upperϥեåȤ֤򤺤餷ƽŤ͡Ϥ߽Фʬ̵뤵
int overlay_xpm(dynamic_xpm_t *lower_dxpm, dynamic_xpm_t *upper_dxpm, 
                int offset_x, int offset_y,
                dynamic_xpm_t *result_dxpm)
{
    int x, y, ux, uy;
    pixel_str_t upper_str;

    assert(result_dxpm->width_max >= lower_dxpm->width);
    assert(result_dxpm->height_max >= lower_dxpm->height);

    result_dxpm->width = lower_dxpm->width;
    result_dxpm->height = lower_dxpm->height;
    copy_dxpm_color(lower_dxpm, result_dxpm);

    for (y = 0 ; y < lower_dxpm->height ; y++){
        for (x = 0 ; x < lower_dxpm->width ; x++){
            ux = x - offset_x;
            uy = y - offset_y;
            if (ux >= 0 && ux < upper_dxpm->width &&
                uy >= 0 && uy < upper_dxpm->height)
            {
                upper_str = DXpmPixel(upper_dxpm, ux, uy);
                if (!IsSamePixelStr(upper_str, upper_dxpm->transparent_str)){
                    DXpmPixel(result_dxpm, x, y) = upper_str;
                    continue;
                }
            }
            DXpmPixel(result_dxpm, x, y) = DXpmPixel(lower_dxpm, x, y);
        }
    }
    return 0;
}

// src_dxpmĲ줾xscale, yscaleܤdest_dxpm˺
// ܤǤʤФʤʤ
int magnify_xpm_int(int xscale, int yscale,
                    dynamic_xpm_t *src_dxpm, dynamic_xpm_t *dest_dxpm)
{
    int x, y, bx, by;

    assert(xscale != 0);
    assert(yscale != 0);
    assert(dest_dxpm->width_max >= src_dxpm->width * xscale);
    assert(dest_dxpm->height_max >= src_dxpm->height * yscale);

    dest_dxpm->width = src_dxpm->width * xscale;
    dest_dxpm->height = src_dxpm->height * yscale;
    copy_dxpm_color(src_dxpm, dest_dxpm);

    for (y = 0 ; y < src_dxpm->height ; y++){
        for (x = 0 ; x < src_dxpm->width ; x++){
            for (by = 0 ; by < yscale ; by++){
                for (bx = 0 ; bx < xscale ; bx++){
                    DXpmPixel(dest_dxpm, x*xscale+bx, y*yscale+by)
                        = DXpmPixel(src_dxpm, x, y);
                }
            }
        }
    }
    return 0;
}


// src_dxpmĲ줾xscale, yscaleܤdest_dxpm˺
// ϼ¿ǡ1̤Ǥн̾
// γƥԥˤĤơб֤ˤäȤᤤԥѤ
// ʤΤǡϱ®Ŀơ֥뤬Ѳʤ
int magnify_xpm(double xscale, double yscale,
                dynamic_xpm_t *src_dxpm, dynamic_xpm_t *dest_dxpm)
{
    int x, y, sx, sy;

    assert(xscale != 0);
    assert(yscale != 0);
    assert(dest_dxpm->width_max >= (int)(src_dxpm->width * xscale));
    assert(dest_dxpm->height_max >= (int)(src_dxpm->height * yscale));

    dest_dxpm->width = (int)(src_dxpm->width * xscale);
    dest_dxpm->height = (int)(src_dxpm->height * yscale);
    copy_dxpm_color(src_dxpm, dest_dxpm);

    for (y = 0 ; y < dest_dxpm->height ; y++){
        for (x = 0 ; x < dest_dxpm->width ; x++){
            sx = x / xscale; 
            sy = y / yscale; 
            DXpmPixel(dest_dxpm, x, y) = DXpmPixel(src_dxpm, sx, sy);
        }
    }
    return 0;
}



int save_xpm(char *path, char *name, char **xpm){
    FILE *fp;
    int width, height, color_num, depth;
    int i;

    if (sscanf(xpm[0], "%d%d%d%d", &width, &height, &color_num, &depth) != 4){
        return 1;
    }
    fp = fopen(path, "w");
    if (fp == NULL) return 1;
    fprintf(fp, "/* XPM */\n");
    fprintf(fp, "static char *%s[] = {\n", name);
    for (i = 0 ; i < 1+color_num+height ; i++){
        fprintf(fp, "\"%s\",\n", xpm[i]);
    }
    fprintf(fp, "};\n");
    fclose(fp);
    return 0;
}





#ifdef DEBUG
void show_dynamic_xpm(dynamic_xpm_t *dxpm){
    int x, y;
    printf("%dx%d\n %d cols\n", dxpm->width, dxpm->height, dxpm->color_num);
    for (y = 0 ; y < dxpm->height ; y++){
        for (x = 0 ; x < dxpm->width ; x++){
            printf("%c%c", 
                   PixelStr(DXpmPixel(dxpm, x, y), 0), 
                   PixelStr(DXpmPixel(dxpm, x, y), 1) 
                   );
        }
        puts("");
    }
}
int show_xpm(char **xpm){
    int width, height, color_num, depth;
    int i;

    if (sscanf(xpm[0], "%d%d%d%d", &width, &height, &color_num, &depth) != 4){
        return 1;
    }
    for (i = 0 ; i < 1+color_num+height ; i++){
        printf("%s\n", xpm[i]);
    }
    return 0;
}
#endif

