#include <stdio.h>
#include <stdlib.h>
#include <EzGraph.h>
#include "config.h"
#include "sprite.h"
#include "obj.h"
#include "param.h"
#include "coordinate.h"
#include "process_xpm.h"

#include "xpm_data.h"

#define BASE_SIZE 24
//#define IMG_SIZE_MAX 120
//#define IMG_SIZE_MAX 152
//#define IMG_SIZE_MAX 184
#define IMG_SIZE_MAX 560
#define XPM_ORG_SIZE_MAX 240
#define COLOR_NUM_MAX 256


//EzImage dummyimg;
//#define EzReadXPM(x) dummyimg

//static int img_sizes[] = {12, 24, 36, 48, 60, 72, 84, 96, 108, IMG_SIZE_MAX};
//static int img_sizes[] = {4, 8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 60, 64, 68, 72, 76, 84, 96, 108, 120, 136, IMG_SIZE_MAX};
static int img_sizes[] = {4, 8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 60, 64, 68, 72, 76, 84, 96, 108, 120, 136, 152, 168, 184, 200, 224, 248, 272, 296, 320, 360, 400, 440, 480, 520, IMG_SIZE_MAX};
#define IMG_SIZE_NUM (sizeof(img_sizes)/sizeof(int))

typedef struct {
    obj_type_t oid;
    sprite_id_t sid;
    char **xpm[2];
    int size;
    int size_min, size_max;
} xpm_resource_t;

#define XR(name, size)                                             \
    {ot_ ## name, si_ ## name, {name ## _f_xpm, name ## _b_xpm},    \
            size, 0, IMG_SIZE_MAX}
/*
#define XRR(name, size, minsize, maxsize)                         \
    {ot_ ## name, si_ ## name, {name ## _f_xpm, name ## _b_xpm},    \
            size, minsize, maxsize}
*/
#define XRrobo(name, size) \
    {ot_myrobo, si_myrobo_ ## name, \
            {myrobo_f_ ## name ## _xpm, myrobo_b_ ## name ## _xpm}, \
            size, 96, 96}
#define XRdrone(name, fn, bn)                                  \
    {ot_drone, si_drone_ ## name,                              \
            {drone_f_ ## fn ## _xpm, drone_b_ ## bn ## _xpm}, \
            64, 0, IMG_SIZE_MAX}

static xpm_resource_t xpm_resources[] = {
    //    XR(chee, 32),
    //    XR(chee, 32),
    //    {ot_myrobo, si_chee, {chee_f_xpm, chee_b_xpm}, 64},
    {ot_myrobo, si_myrobo, {myrobo_f_xpm, myrobo_b_xpm}, 96, 96, 96},
    {ot_myrobo, si_myrobo_title, {myrobo_f_title_xpm, myrobo_b_title_xpm},
     96, 96, 96},

    XRrobo(turn_l1, 96),
    XRrobo(turn_l2, 96),
    XRrobo(turn_l3, 96),
    XRrobo(turn_l4, 96),
    XRrobo(turn_l5, 96),
    XRrobo(turn_l6, 96),
    XRrobo(turn_l7, 96),
    XRrobo(turn_l8, 96),
    XRrobo(turn_l9, 96),
    XRrobo(turn_r1, 96),
    XRrobo(turn_r2, 96),
    XRrobo(turn_r3, 96),
    XRrobo(turn_r4, 96),
    XRrobo(turn_r5, 96),
    XRrobo(turn_r6, 96),
    XRrobo(turn_r7, 96),
    XRrobo(turn_r8, 96),
    XRrobo(turn_r9, 96),

    XR(dummy_target, 64),
    XR(enemy_destroyer, 160),

    XRdrone(00, 00, 08),
    XRdrone(01, 15, 07),
    XRdrone(02, 14, 06),
    XRdrone(03, 13, 05),
    XRdrone(04, 12, 04),
    XRdrone(05, 11, 03),
    XRdrone(06, 10, 02),
    XRdrone(07, 09, 01),
    XRdrone(08, 08, 00),
    XRdrone(09, 07, 15),
    XRdrone(10, 06, 14),
    XRdrone(11, 05, 13),
    XRdrone(12, 04, 12),
    XRdrone(13, 03, 11),
    XRdrone(14, 02, 10),
    XRdrone(15, 01, 09),
//    XR(chee, 64),
    //    XR(squid, 24),
    XR(asteroid, 64),
    XR(mine, 64),
    {ot_asteroid, si_explosion, {explosion_xpm, explosion_xpm}, 64, 0, IMG_SIZE_MAX},
    //    XR(watermelon, 24),
};
#define XPM_RESOURCE_NUM (sizeof(xpm_resources)/sizeof(xpm_resource_t))

typedef struct {
    int base_size;
    EzImage imgs[IMG_SIZE_NUM][2];
} sprite_t;

static sprite_t sprites[si_end];

int get_sprite_img(int size, dir_t dir, sprite_id_t sid, EzImage *img);

static int img_ix_table[IMG_SIZE_MAX+1];

static int obj_box_enabled = OBJ_BOX_ENABLED_DEFAULT;


int sprite_init(void){
    int i, j, k, img_ix, size_limit, min_size_no, max_size_no;
    dynamic_xpm_t *src_dxpm, *dest_dxpm;
    char **xpm_buf;
    double scale;
    int img_num = 0;
    char buf[256];

    img_ix = 0;
    for (i = 0 ; i <= IMG_SIZE_MAX ; i++){
        if (i > img_sizes[img_ix]){
            img_ix ++;
        }
        img_ix_table[i] = img_ix;
    }

    src_dxpm = alloc_dynamic_xpm(XPM_ORG_SIZE_MAX, XPM_ORG_SIZE_MAX);
    dest_dxpm = alloc_dynamic_xpm(IMG_SIZE_MAX, IMG_SIZE_MAX);
    xpm_buf = alloc_xpm_buffer(IMG_SIZE_MAX, IMG_SIZE_MAX, COLOR_NUM_MAX);
    
    for (i = 0 ; i < XPM_RESOURCE_NUM ; i++){
        size_limit = Size2ScreenSize(get_obj_size(xpm_resources[i].oid),
                                     SCREEN_RZMIN);
        if (size_limit > xpm_resources[i].size_max){
            size_limit = xpm_resources[i].size_max;
        }
        min_size_no = 0;
        max_size_no = -1;
        for (j = 0 ; j < IMG_SIZE_NUM ; j++){
            if (img_sizes[j] <= xpm_resources[i].size_min){
                min_size_no = j;
            }
            if (img_sizes[j] > size_limit){
                max_size_no = j;
                break;
            }
        }
        if (max_size_no < 0){
            max_size_no = IMG_SIZE_NUM-1;
        }
        sprintf(buf, "Generating 3D Sprite %3d/%3d", i, XPM_RESOURCE_NUM);
        EzDrawStringB(80, 100, buf);
        EzShowBuffer();
        printf("Generating Sprite id:%d Size:%d(limit %d-%d) img_num:%d\n", i, img_sizes[j], xpm_resources[i].size_min, size_limit, img_num);
        for (j = min_size_no ; j <= max_size_no ; j++){
            //            printf("Generationg Sprite id:%d Size:%d(limit %d-%d)\n", i, img_sizes[j], xpm_resources[i].size_min, size_limit);
            for (k = 0 ; k < 2 ; k++){
                xpm2dynamic_xpm(xpm_resources[i].xpm[k], src_dxpm);
                //            if (img_sizes[j] == BASE_SIZE){
                if (img_sizes[j] == xpm_resources[i].size){
                    sprites[xpm_resources[i].sid].imgs[j][k]
                        = EzReadXPM(xpm_resources[i].xpm[k]);
                    img_num ++;
                }
                else {
                    //                scale = (double)img_sizes[j] / BASE_SIZE;
                    scale = (double)img_sizes[j] / xpm_resources[i].size;
                    
                    magnify_xpm(scale, scale, src_dxpm, dest_dxpm);
                    dynamic_xpm2xpm(dest_dxpm, xpm_buf);
                    sprites[xpm_resources[i].sid].imgs[j][k]
                        = EzReadXPM(xpm_buf);
                    img_num ++;
                }
            }
        }
    }
    return 0;
}

void draw_sprite(int x, int y, int size, dir_t dir, sprite_id_t sid){
    EzImage img;
    int offset_x, offset_y;
    int is_invalid;

    is_invalid = get_sprite_img(size, dir, sid, &img);
    offset_x = img.width/2;
    offset_y = img.height/2;
    //    if (sid == si_enemy_destroyer) printf("size:%d ssize:%d sid:%d valid:%d mg%dx%d \n", size, img.width, is_invalid, sid, img.width, img.height);

    if (is_invalid){
        if (obj_box_enabled){
            EzSetColorByRGB(255, 0, 0);
            EzDrawBoxB(x-offset_x, y-offset_y, x+offset_x-1, y+offset_y-1);
        }
    }
    else {
        if (obj_box_enabled){
            EzSetColorByRGB(0, 255, 0);
            EzDrawBoxB(x-offset_x, y-offset_y, x+offset_x-1, y+offset_y-1);
        }
        //        EzPutB(x-sprite_size/2, y-sprite_size/2, img);
        EzPutB(x-offset_x, y-offset_y, img);
    }
}

int get_sprite_img(int size, dir_t dir, sprite_id_t sid, EzImage *img){
    int img_ix;

    if (size > IMG_SIZE_MAX) return 1;
    img_ix = img_ix_table[size];
    //    img_size.x = img_sizes[img_ix];
    //    printf("size:%d->%d(ix:%d)\n", size, img_size, img_ix);
    /*
    img_ix = size/24;
    if (img_ix >= IMG_SIZE_NUM) return 0;
    */

    *img = sprites[sid].imgs[img_ix][dir];
#ifdef DEBUG
    if (img->width == 0){
        printf("illegal image size:%d(%d) sprite id:%d\n", img->width, size, sid);
        exit(1);
    }
#endif
    return 0;
}
